;; A function that makes a copy of a list of constant immutable
;; objects.
;;
(define (copy-simple-list lst)
   (if (not (list? lst))
       (begin
	  (er "copy-simple-list" "the argument is not a list")
	  '*error-copy-simple-list*)
       (let loop ((l lst)(res '()))
	  (if (null? l)
	      res
	      (loop (cdr l) (append res (list (car l))))))))

;; A function that makes a copy of a vector of constant immutable
;; objects.
;;
(define (copy-simple-vector vec)
   (if (not (vector? vec))
       (begin
	  (er "copy-simple-vector" "the argument is not a vector")
	  '*error-copy-simple-vector*)
       (list->vector (vector->list vec))))

;; A function that makes a copy of a loopy matrix.
;;
(define (copy-matrix mat)
   (if (not (matrix? mat))
       (begin
	  (er "copy-matrix" "the argument is not a matrix")
	  '*error-copy-matrix*)
       (make-matrix (copy-simple-vector (matrix-dims mat))
		    (copy-simple-vector (matrix-data mat)))))

;; A function that makes a copy of a variable node structure.
;; Note that the function does not copy the list of connections.
;; 
(define (copy-variable-node var)
   (if (not (variable-node? var))
       (begin
	  (er "copy-variabe-node" "the argument is not a variable node")
	  '*error-copy-variable-node*)
       (make-variable-node (string-copy (variable-node-id var))
			   (copy-simple-list (variable-node-term var))
			   (copy-simple-list (variable-node-tick var))
			   (variable-node-dim var)
			   (copy-simple-vector (variable-node-belief var))
			   (list))))

;; A function that makes a copy of a cluster node. Note that the function
;; does not copy node's connections including connections to a learnable node.
;;
(define (copy-cluster-node clust)
   (if (not (cluster-node? clust))
       (begin
	  (er "copy-cluster-node" "the argument is not a cluster node")
	  '*error-copy-cluster-node*)
       (let ((learn (cluster-node-learnable clust)))
	  (if (null? l)
	      ;; if the cluster node is not connected to a learnable node, we just copy its table while copying the node
	      (make-cluster-node (string-copy (cluster-node-id clust))
				 (copy-matrix (cluster-node-table clust))
				 (copy-matrix (cluster-node-potential clust))
				 (list)
				 (list))
	      ;; otherwise we take the current approximation from the learnable node and copy it as a table of the cluster node
	      (make-cluster-node (string-copy (cluster-node-id clust))
				 (copy-matrix (conditionally-normalize-matrix (learnable-node-table (car l))))
				 (copy-matrix (cluster-node-potential clust))
				 (list)
				 (list))))))



(define (complete-query-net-from-query q curent-net query-net)
   ;; try to find a variable node correcponding to a term
   (let* ((qterm (get-loopy-query-term q))(qvar (get-variable-node qterm (manager-network man))))
      (if (not qvar)
	  ;; if no such variable node is found, create a new one
	  (let ((new-var (create-new-variable-node qterm
						   (get-loopy-definition-dim (definitions-hash 'get-first-entry (car qterm))))))
	     (add-node-to-network (manager-network man) new-var)
	     (add-node-to-frame man new-var)
	     ;; complete the network for the new variable node
	     (complete-net-from-vars (list new-var) man)))))

(define (make-query-net current-time current-net) ;; current-time == real-current-time
   (define (convert-to-number term)
      (cond ((eq? term (string->symbol (number->string 0))) 0)
	    ((and (list? term) (not (null? term)) (eq? (car term) 'succ)) (+ 1 (convert-to-number (cadr term))))
	    (else (er "convert-to-number" "the argument is empty")
		  '*error-argument*)))

   (define (make-relative-query q t)
      (let* ((qterm (query-term q))
	     (qname (car qterm))
	     (qtime (convert-to-number (cadr qterm)))
	     (time (convert-to-number t))
	     (diff (- qtime time)))
	 (cond ((> diff 0) (let loop ((n diff)(res 't))
			      (if (= n 0)
				  (query-term-set! q (cons qname (list res)))
				  (loop (- n 1) (cons 'succ (list res))))))
	       ((< diff 0) (let loop ((n diff)(res 't))
			      (if (= n 0)
				  (query-term-set! q (cons qname (list res)))
				  (loop (+ n 1) (cons 'pred (list res))))))
	       (else (query-term-set! q (cons qname (list 't)))))))
   
   (let ((query-net (create-new-network (list) (list) (list))))
      (let loop ((l queries-list))
	 (if (null? l)
	     query-net
	     (let ((elem (make-relative-query (car l) current-time)))
;		(complete-query-net-from-query elem current-net query-net);;???? do I need to assign?
		(loop (cdr l)))))))

