(define (attach-leaf fact var man)
   (cond ((not (fact? fact)) (er "attach-leaf" "the first argument is not a fact")
			     '*error-argument*)
	 ((not (variable-node? var)) (er "attach-leaf" "the second argument is not a variable node")
				     '*error-argument*)
	 ((not (manager? man)) (er "attach-leaf" "the third argument is not a manager")
			       '*error-argument*)
	 (else (let ((clust (create-new-cluster-node (list var) (get-loopy-fact-value fact))))
		  (add-node-to-network (manager-network man) clust)
		  (add-fact-to-frame (get-appropriate-frame var man) fact)))))

(define (get-observable var facts)
   (display "GET-OBSERVABLE\n");;;
   (display "VAR: ")(pretty-print (variable-node-term var));;;
   (display "FACTS: ")(pretty-print facts);;;
   (cond ((not (variable-node? var)) (er "get-observable" "the first argument is not a variable node")
				     '*error-argument*)
	 (else (let ((term (get-variable-node-term var)))
		  (let loop ((l facts)(prefix '()))
		     (if (null? l)
			 (cons #f facts)
			 (let ((fact (car l)))
			    (if (equal? term (fact-term fact))
				(cons fact (append prefix (cdr l)))
				(loop (cdr l) (append prefix (list fact)))))))))))

(define (complete-failure-net-from-vars lst facts man)
   (display "complete-failure-net-from-vars\n")
   (display "VARS: ")(pretty-print (map variable-node-term lst));;;;
   (display "FACTS: ")(pretty-print facts);;;;
   (cond ((not (manager? man)) (er "complete-failure-net-from-vars" "the last argument is not a manager")
			       '*error-argument*)
	 (else (let loop ((vars lst)(facts facts))
		  (display "LOOP\n");;;
		  (display "VARS: ")(pretty-print (map variable-node-term vars));;;;
		  (display "FACTS: ")(pretty-print facts);;;;
		  (if (not (null? vars))
		      (let* ((var (car vars))
			     (ret (get-observable var facts))
			     (fact (car ret))
			     (remaining-facts (cdr ret)))
			 (display "VAR: ")(pretty-print (variable-node-term var));;;
			 (display "FACT: ")(pretty-print fact);;;
			 (display "REMAINING-FACTS: ")(pretty-print remaining-facts);;;
			 (if fact
			     (begin
				(display "ATTACH\n")
				(attach-leaf fact var man)
				(loop (cdr vars) remaining-facts))
			     (let iter ((rules (rules-hash 'get-every-entry (car (get-variable-node-term var))))
					(neighbors (cdr vars)))
				(display "ITER\n")
				(display "RULES(HEADS): ")(pretty-print (map rule-head rules));;;;
				(display "NEIGHBORS: ")(pretty-print (map variable-node-term neighbors));;;;
				(if (or (null? rules) (eq? rules '*hash-table-search-failed*))
				    (loop neighbors facts)
				    (begin
				       (display "VAR: ")(pretty-print (variable-node-term var));;
				       (display "RULE(HEAD): ")(pretty-print (rule-head (car rules)));
				       (display "CONSTR-NET-FROM-RULE\n");
				       (let ((res (construct-subnet-from-rule var (car rules) man)))
					  (display "RES: ")(pretty-print (map variable-node-term res));;
					  (if (eq? res '*error-construct-network*)
					      '*error-construct-network*
					      ;; add new variable nodes on the list without repetition
					      (iter (cdr rules) (merge-lists neighbors res))))))))))))))

(define (expand-failure-network-one-time-step man dat rule)
   (define (get-next-tick tick)
      (cons 'succ (list tick)))

   (define (time-of-fact? f t)
      (cond ((not (fact? f))
	     (er "time-of-fact?" "the first argument is not a fact")
	     #f)
	    ((null? t)
	     (er "time-of-fact?" "the second argument is not a time tick")
	     #f)
	    (else (equal? (cadr (fact-term f)) t))))
   
   (define (make-fact-time-relative f)
      (let ((term (fact-term f)))
	 (fact-term-set! f (list (car term) 't))))

   (cond ((not (manager? man))
	  (er "expand-network-one-time-step" "the first argument is not a manager")
	  '*error-argument*)
	 ((not (data-control? dat)) 
	  (er "expand-network-one-time-step" "the second argument is not a data control")
	  '*error-argument*)
	 ((not (rule? rule)) 
	  (er "expand-network-one-time-step" "the third argument is not a rule")
	  '*error-argument*)
	 (else
	  (let* ((facts (get-current-facts dat))
		 (term (cons (car (get-loopy-rule-head rule))(list 't)))
		 (var (create-new-variable-node term (get-loopy-definition-dim (definitions-hash 'get-first-entry (car term))))))
	     (add-node-to-network (manager-network man) var)
	     (add-node-to-frame man var)
	     (display "MADE FIRST: ") (pretty-print (variable-node-term var));;;;
	     ;; for the facts of a time step, make then relative
	     (let loop ((l facts))
		(if (not (null? l))
		    (begin
		       (make-fact-time-relative (car l))
		       (loop (cdr l)))))
	     (display "MADE FACTS RELATIVE\n");;;;
	     (let ((new-vars (construct-subnet-from-rule var rule man)))
		(display "MADE RULE NET\n");;;
		(display "NEW VARS: ") (pretty-print (map variable-node-term new-vars));;;;
		(let* ((ret (get-observable var facts))
		       (fact (car ret))
		       (remaining-facts (cdr ret)))
		   (display "FACT: ")(pretty-print fact);;
		   (display "REMAINING-FACTS: ")(pretty-print remaining-facts);;
		   (if fact
		       (begin
			  (display "ATTACH FACT\n")
			  (attach-leaf fact var man)
			  (display "COMPLETE NET FOR VARS\n")
			  (complete-failure-net-from-vars new-vars remaining-facts man))
		       (let loop ((rules (rules-hash 'get-every-entry (car term)))
				  (neighbors new-vars))
			  (if (or (null? rules) (eq? rules '*hash-table-search-failed*))
			      (begin
				 (display "COMPLETE NET FOR VARS\n")
				 (complete-failure-net-from-vars neighbors facts man)
				 )
			      (let ((r (car rules)))
				 (display "RULE(HEAD): ")(pretty-print (rule-head r));;
				 (if (not (eq? r rule))
				     (let ((res (construct-subnet-from-rule var r man)))
					(if (eq? res '*error-construct-network*)
					    '*error-construct-network*
					    (loop (cdr rules) (merge-lists neighbors res))))
				     (loop (cdr rules) neighbors))))))))))))


(define (complete-failure-network-for-rule man dat rule)
   
   (define (convert-to-number term)
      (cond ((eq? term (string->symbol (number->string 0))) 0)
	    ((and (list? term) (not (null? term)) (eq? (car term) 'succ)) (+ 1 (convert-to-number (cadr term))))
	    (else (er "convert-to-number" "the argument is empty")
		  '*error-argument*)))

   (cond ((not (manager? man))
	  (er "complete-network-for-rule" "the first argument is not a manager")
	  '*error-argument*)
	 ((not (data-control? dat)) 
	  (er "complete-network-for-rule" "the second argument is not a data control")
	  '*error-argument*)
	 ((not (rule? rule)) 
	  (er "complete-network-for-rule" "the third argument is not a rule")
	  '*error-argument*)
	 (else
	  (if (< (length (manager-frames man))
		 (manager-window_size man))
	      (if (null? (data-control-given_facts dat))
		  (let ((new-facts (get-next-facts dat)))
		     (if (not (null? new-facts))
			 (let ((gap (car new-facts)))
			   (let iter ((i gap))
			       (if (> i 0)
				   (begin
				      (shift-time-frames man)
				      (iter (- i 1)))))
			    (display "HERE 1\n")
			    (expand-failure-network-one-time-step man dat rule)
			    (display "HERE 2\n")
			    (complete-failure-network-for-rule man dat rule))))
		  (begin
		     (get-next-given-facts dat)
		     (let ((gap (get-current-gap dat)))
		       (let iter ((i gap))
			 (if (> i 0)
			     (begin
			       (shift-time-frames man)
			       (iter (- i 1)))))
		       (display "HERE 3\n")
		       (expand-failure-network-one-time-step man dat rule)
		       (display "HERE 4\n")
		       (complete-failure-network-for-rule man dat rule))))))))
   



(define (initialize-rule-manager r)
   (display "INITIALIZATION (hacked window size): ")(pretty-print r);;
   (let ((man (create-manager #f #f #f window-size)))
;   (let ((man (create-manager #f #f #f 8)))
      (display "MADE EMPTY MANAGER\n")
      (rule-failure_manager-set! r man)))


(define (convert-real-matrix mat)

  (define (list-mult l)
    (let loop ((l l)(res 1))
      (cond ((null? l) res)
	    (else (loop (cdr l) (* (car l) res))))))
  
  (define (split l n)
    (cond ((>= n (length l)) (list l (list)))
	  ((zero? n) (list (list) l))
	  (else (let ((s (split (cdr l) (- n 1))))
		  (list (cons (car l) (car s)) (car (cdr s)))))))
  
  (define (reshuffle data dim)
    (if (= (length dim) 1)
	data
	(let ((i (car dim))(j (list-mult (cdr dim)))) 
	  (let loop ((iter 0)
		     (shift (- j 1))
		     (elem (list))
		     (new-data (list))
		     (split-list (list))
		     (l data))
	    (cond ((< iter i) (let ((pair (split (cdr l) shift)))
				(loop (+ iter 1)
				      shift
				      (append elem (list (car l))) 
				      (append new-data (car pair)) 
				      split-list
				      (car (cdr pair)))))
		  ((> shift 0) (loop 0 
				     (- shift 1) 
				     (list) 
				     (list) 
				     (append split-list (list elem)) 
				     new-data))
		  (else (reshuffle (append split-list (list elem)) (cdr dim))))))))

  (define (convert data)
    (define (zero-list? l)
      (cond ((null? l) #t)
	    ((zero? (car l)) (zero-list? (cdr l)))
	    (else #f)))

    (define (convert-numbers data)
      (if (null? data)
	  (display "empty")
	  (let* ((z (zero-list? data))(len (length data))(elem (/ 1.0 len)))
	    (let loop ((l data)(res (list)))
	      (cond ((null? l) res)
		    ((zero? (car l)) (if z
					 (loop (cdr l) (append res (list elem)))
					 (loop (cdr l) (append res (list (string->number "0.1e-250"))))))
		    (else (loop (cdr l) (append res (list (car l))))))))))
      
    (cond ((null? data) (list))
	  ((list? (car data)) (cons (convert (car data)) (convert (cdr data))))
	  (else (convert-numbers data))))

  (if (loopy-matrix? mat)
      (let ((dim (get-matrix-dimensions mat))(data (get-matrix-data mat)))
	(let ((cont-data (reshuffle (vector->list data) (vector->list dim))))
	  (convert cont-data)
	  ))
      (begin 
	(er "convert-real-matrix" "the argument is not a matrix")
	#f)))

(define (list-distance v1 v2)
   (cond
    ((not (list? v1)) (begin
			(er "list-distance" "the first argument is not a list")
			-1))
    ((not (list? v2)) (begin
			(er "list-distance" "the second argument is not a list")
			-1))
    (else (cond ((null? v1) 0.0)
		((list? (car v1)) (+ (list-distance (car v1) (car v2)) (list-distance (cdr v1) (cdr v2))))
		(else (let loop ((l1 v1)(l2 v2)(acc 0.0))
			(if (null? l1)
			    acc
			    (let ((dif (- (car l1) (car l2))))
			      (loop (cdr l1) (cdr l2) (+ acc (* dif dif)))))))))))



(define (check-for-failure dat)

   (define (change-learnables r)
      (let* ((net (manager-network (rule-failure_manager r)))
	     (learn-list (network-learnable net)))
	 (let loop ((l learn-list)(acc #f))
	    (if (null? l)
		(network-learnable-set! net (list acc))
		(let ((learn (car l)))
		   (if (eq? (learnable-node-loopy_rule learn) r)
		       (loop (cdr l) learn)
		       (let iter ((clusts (learnable-node-neighbors learn)))
			  (if (null? clusts)
			      (loop (cdr l) acc)
			      (let* ((c (car clusts))(cpt (rule-learned (learnable-node-loopy_rule learn))))
				 (cluster-node-table-set! c cpt)
				 (cluster-node-learnable-set! c '())
				 (iter (cdr clusts)))))))))))
   
   (display "CHECK-FOR-FAILURE\n");;

;; this is a hack to add an initial bias to the hmm
;   (let ((fff (get-next-facts dat));;;;;;;;;;;;;
;	 (newf (make-fact (list 'state 't) ;;;;;;;;;;;;;;;;;;;;
;			  (make-matrix (vector 2) (vector 1 0)))));;;;;;;;;;;;;;;;;;;
;     (display "FFFFFF: ")(pretty-print fff)
;     (display "NNNNNN: ")(pretty-print newf)
;     (data-control-current_facts-set! dat (append fff (list newf)));;;;;;;;;;;;;;
;     (reuse-previously-seen-facts dat));;;;;;;;;;;;;;

   (let loop ((rules (rules-hash 'get-all-entries)))
      (display "MAIN LOOP\n")
      (if (null? rules)
	  #f
	  (let ((current-rule (cdar rules))
		(learning #t))
	     (display "CURRENT-RULE: ")(pretty-print (rule-head current-rule));;

	     (let ((mat (rule-matrix current-rule)))
	       (if (loopy-matrix? mat)
		   (begin
		     (rule-learned-set! current-rule mat)
		     (rule-matrix-set! current-rule 'Temp)
		     (set! learning #f))))

	     (if (not (rule-failure_manager current-rule))
		    (initialize-rule-manager current-rule))
	     (display "INITIALIZED CURRENT-RULE: ")(pretty-print (rule-head current-rule));;
	     (display "WSIZE: ")(pretty-print (manager-window_size (rule-failure_manager current-rule)));;;
	     (display "COMPLETE-NET\n");;
;	     (complete-failure-network-for-rule (rule-failure_manager current-rule)
;						dat
;						current-rule)
	     (complete-network-from-file (rule-failure_manager current-rule) dat 'complete)
	     (display "CHANGING LEARNABLES\n");;
	     (change-learnables current-rule)
	     (display "FAILURE-NET\n")(print-network (manager-network (rule-failure_manager current-rule)));;
	     (let ((iter (run-update (manager-network (rule-failure_manager current-rule)) 99 0.000001)))
		(if (eq? iter '*error-run-update*)
		    (er "check-for-failure" "there was an error during inferencing over a failure network")
		    (if trace_flag
			(begin
			   (display "UPDATED NETWORK:\n\n")
			   (print-network (manager-network (rule-failure_manager current-rule)))))))
	     (let ((new-learned (learnable-node-table (car (network-learnable (manager-network (rule-failure_manager current-rule))))))
		   (learned (rule-learned current-rule)))
		(let* ((new-learn (convert-real-matrix (conditionally-normalize-matrix new-learned)))
		       (old-learn (convert-real-matrix learned))
		       (dist (list-distance new-learn
					    old-learn)))
		   (display "OLD-LEARN: ")(pretty-print old-learn);;
		   (display "NEW-LEARN: ")(pretty-print new-learn);;
		   (display "DIST: ")(pretty-print dist);;
		   (display "RULE: ")(pretty-print (rule-head current-rule));;

		   (if (not learning)
		       (begin
			 (rule-matrix-set! current-rule learned)
			 (rule-learned-set! current-rule #f)))

		   (if (< 100 dist)
		       current-rule
		       (begin
			  (pretty-print dat)
			  (reuse-previously-seen-facts dat)
			  (pretty-print dat)
			  (loop (cdr rules))))))))))
