
;; the definition of a frame structure
;; time_tick       -- a time tick of a frame in relative time    (list symbol)
;; variable_nodes  -- a list of variable nodes                   (list variable-node)
;; facts           -- a list of facts                            (list fact)

(define-structure frame time_tick variable_nodes facts)

;; A function that creates a new frame and returns it.
;; Note that the new frame corresponding
;; to the current time step has a time tick equal to "t".

(define (create-new-frame tick)
   (make-frame tick (list) (list)))


;; A function that adds a fact to the list of a frame
;;
(define (add-fact-to-frame fr f)
   (cond ((not (frame? fr))
	  (er "add-fact-to-frame" "the first argument is not a frame")
	  '*error-argument*)
	 ((not (fact? f))
	  (er "add-fact-to-frame" "the second argument is not a fact")
	  '*error-argument*)
	 (else
	  (let ((facts (frame-facts fr)))
	     (frame-facts-set! fr (append facts (list f)))))))

;; A function that adds a variable node to a frame and returns a new
;; list of variables of the frame.

(define (add-variable-node-to-frame! fr node)
   (cond ((not (frame? fr))
	  (er "add-variable-node-to-frame!" "the first argument is not a frame")
	  '*error-argument*)
	 ((not (variable-node? node))
	  (er "add-variable-node-to-frame!" "the second argument is not a variable node")
	  '*error-argument*)
	 (else (let ((node-list (append (frame-variable_nodes fr) (list node))))
		  (frame-variable_nodes-set! fr node-list)
		  node-list))))

;; A function that finds an appropriate frame (or creates a new frame),
;; adds a node to the frame, and returns the frame.
;;
(define (add-node-to-frame man node)
   (if (variable-node? node)
       (let ((fr (get-appropriate-frame node man)))
	  (add-variable-node-to-frame! fr node)
	  fr)
       (begin
	  (er "add-node-to-frame" "the argument is not a variable node")
	  '*error-arguments*)))


;; A function that prints a frame. Note that it only prints out the IDs of
;; nodes in the lists.

(define (print-frame fr)
   (if (frame? fr)
       (begin
	  (display "+++++++ FRAME +++++++\n\n")
	  (display "Time Tick:  ")
	  (pretty-print (collapse-time (frame-time_tick fr)))
	  (display "\n")
	  (display "Variable Nodes:  ")
	  (pretty-print (map get-variable-node-id (frame-variable_nodes fr)))
	  (display "\n")
	  (display "Facts:  ")
	  (pretty-print (frame-facts fr))
	  (display "+++++++ +++++ +++++++\n\n"))
       (begin
	  (er "print-frame" "the argument is not a frame")
	  '*error-argument*)))

