(define (visualize-vector vect output)
   (let loop ((l (vector->list vect)))
      (if (not (null? l))
	  (let ((elem (car l)))
	     (display (/ (round (* elem 1000000)) 1000000) output)
	     (display " " output)
	     (loop (cdr l))))))

(define (visualize-term term output)
   (display "(" output)
   (let loop ((l term))
      (if (null? l)
	  (display ")" output)
	  (let ((elem (car l)))
	     (if (list? elem)
		 (visualize-term elem output)
		 (begin
		    (display elem output)
		    (if (not (null? (cdr l)))
			(display " " output))))
	     (loop (cdr l))))))

(define (visualize-variable var output)
   (let ((id (get-variable-node-id var))
	 (term (get-variable-node-term var)))
      (display "node [shape=ellipse,style=filled,color=khaki,label=\"" output)
      (visualize-term term output)
      (display ", " output)
      (display id output)
      (display "\"]; " output)
      (display id output)
      (display ";" output)
      (newline output)))

(define (visualize-learn learn output)
   (let ((id (get-learnable-node-id learn)))
      (display "node [shape=triangle,style=filled,color=lightblue,label=\"" output)
      (display id output)
      (display "\"]; " output)
      (display id output)
      (display ";" output)
      (newline output)))

(define (visualize-cluster clust var-neighbors output)
   (let ((id (get-cluster-node-id clust))
	 (table (get-matrix-data (get-cluster-node-table clust)))
	 (potential (get-matrix-data (get-cluster-node-potential clust))))
      (display "node [shape=box,style=filled,color=lightcoral,label=\"" output)
      (display id output)
      (display "\\n" output)
      (visualize-vector table output)
      (display "\\n" output)
      (visualize-vector potential output)
      (display "\"]; " output)
      (display id output)
      (display ";" output)
      (newline output)
      (let loop ((l var-neighbors))
	 (if (not (null? l))
	     (begin
		(display (get-variable-node-id (car l)) output)
		(display " -- " output)
		(display id output)
		(display ";" output)
		(newline output)
		(loop (cdr l)))))))


(define (visualize-subnet cluster output)
   (define (subtract-lists minuend subtrahend)
      (let loop ((l minuend)(res (list)))
	 (if (null? l)
	     (reverse res)
	     (let ((head (car l))(tail (cdr l)))
		(if (memv head subtrahend)
		    (loop tail res)
		    (loop tail (cons head res)))))))
   
   (define (visualize-cluster-neighbors var-neighbors seen-variables output)
      (let loop ((l var-neighbors)(visited (list)))
	 (if (null? l)
	     (reverse visited)
	     (let ((head (car l))(tail (cdr l)))
		(if (memv head seen-variables)
		    (loop tail visited)
		    (begin
		       (visualize-variable head output)
		       (loop tail (cons head visited))))))))

   (let ((seen-variables (list))(seen-clusters (list)))
      (let ((var-neighbors (map get-edge-variable (get-cluster-node-neighbors cluster))))
	 (let loop ((l var-neighbors))
	    (if (not (null? l))
		(let ((head (car l))(tail (cdr l)))
		   (visualize-variable head output)
		   (set! seen-variables (append seen-variables (list head)))
		   (loop tail))))
	 (visualize-cluster cluster var-neighbors output))
      (set! seen-clusters (append seen-clusters (list cluster)))
      (let loop ((l seen-variables))
	 (if (null? l)
	     seen-clusters
	     (let ((not-seen-clusters (subtract-lists (map get-edge-cluster (get-variable-node-neighbors (car l)))
						      seen-clusters)))
		(let next-cluster ((iter not-seen-clusters))
		   (if (null? iter)
		       (loop (cdr l))
		       (let ((var-neighbors (map get-edge-variable (get-cluster-node-neighbors (car iter)))))
			  (let ((visited-vars (visualize-cluster-neighbors var-neighbors seen-variables output)))
			     (set! l (append l visited-vars))
			     (visualize-cluster (car iter) var-neighbors output)
			     (set! seen-clusters (append seen-clusters (list (car iter))))
			     (next-cluster (cdr iter)))))))))))

(define (visualize-net net name)
   (cond ((not (network? net)) (er "visualize-net" "the first argument is not a network")
			       '*error-argument*)
	 ((not (string? name)) (er "visualize-net" "the second argument is not a string")
			       '*error-argument*)
	 (else
	  (let ((output (open-output-file name)))
	     (display "graph ER {" output)
	     (newline output)
	     (display "ranksep=3;" output)
	     (newline output)
	     (display "ratio=auto;" output)
	     (newline output)
	     (let ((learn-list (network-learnable net)))
		(if (null? learn-list)
		    (let ((cluster (car (network-clusters net))))
		       (visualize-subnet cluster output))
		    (let ((learn (car learn-list)))
		       (visualize-learn learn output)
		       (let loop ((l (get-learnable-node-neighbors learn))(seen-clusters (list)))
			  (if (not (null? l))
			      (let ((cluster (car l)))
				 (if (not (memv cluster seen-clusters))
				     (let ((visited-clusters (visualize-subnet cluster output)))
					(set! seen-clusters (append seen-clusters visited-clusters))))
				 (display (get-learnable-node-id learn) output)
				 (display " -- " output)
				 (display (get-cluster-node-id cluster) output)
				 (display ";" output)
				 (newline output)
				 (loop (cdr l) seen-clusters)
				 )))
		       (let loop ((l (cdr learn-list)))
			  (if (not (null? l))
			      (let ((learnable (car l)))
				 (visualize-learn learnable output)
				 (let next-cluster ((clust-list (get-learnable-node-neighbors learnable)))
				    (if (not (null? clust-list))
					(let ((cluster (car clust-list)))
					   (display (get-learnable-node-id learnable) output)
					   (display " -- " output)
					   (display (get-cluster-node-id cluster) output)
					   (display ";" output)
					   (newline output)
					   (next-cluster (cdr clust-list)))))
				 (loop (cdr l))))))))
	     (display "label = \"\\n\\nMarkov network\";" output)
	     (newline output)
	     (display "fontsize=14;" output)
	     (newline output)
	     (display "}" output)
	     (close-output-port output)))))

